import asyncio
import logging
from shutil import copyfile

import click
import pandas as pd

from .cas import CAS
from .check import check_all
from .collect import Collect
from .config import settings as s
from .config import validate_cas
from .config import validate_collect
from .config import validate_db
from .config import validate_s3
from .query import Query
from .templates import secr_template
from .templates import sett_template

logger = logging.getLogger(__name__)


def load_sett_files(settings, secrets):
    if settings:
        logger.info(f"Loading settings file from {settings}")
        s.load_file(path=settings)
    if secrets:
        logger.info(f"Loading secrets file from {secrets}")
        s.load_file(path=secrets)


@click.group()
def cli():
    pass


@cli.command(help="Create template for settings and secrets files")
@click.argument("name")
def init(name):
    logger.info(f"Creating {name}-settings.toml file")
    copyfile(sett_template, f"{name}-settings.toml")
    logger.info(f"Creating {name}.secrets.toml")
    copyfile(secr_template, f"{name}.secrets.toml")
    logger.info("Successful init command")


@cli.command(help="Run a arbitrary query and save results in a .csv")
@click.argument("sql")
@click.option(
    "--name",
    default="result.csv",
    help="Filename for saving results, if not passed default is result.csv",
)
@click.option(
    "--settings",
    default=None,
    help="Path for settings file, if not passed default is settings.toml",
)
@click.option(
    "--secrets",
    default=None,
    help="Path for .secrets file, if not passed default is .secrets.toml",
)
@click.option(
    "--save",
    type=click.Choice(["S3", "local"], case_sensitive=False),
    default="local",
    help="Choose where to save files. If not passed default is local",
)
def query(sql, name, settings, secrets, save):
    load_sett_files(settings, secrets)

    validate_db()
    q = Query(save)
    q.run(sql, name)
    logger.info("Successful query command")


@cli.command(help="Collect files from a database.")
@click.option(
    "--settings",
    default=None,
    help="Path for settings file, if not passed default is settings.toml",
)
@click.option(
    "--secrets",
    default=None,
    help="Path for .secrets file, if not passed default is .secrets.toml",
)
@click.option(
    "--save",
    type=click.Choice(["S3", "local"], case_sensitive=False),
    default="local",
    help="Choose where to save files. If not passed default is local",
)
@click.option(
    "--classes",
    default=None,
    help="CSV with classes in the first column, you can also pass it in settings classes variable.",
)
def collect(settings, secrets, save, classes):
    load_sett_files(settings, secrets)

    validate_db()
    validate_collect()
    if save == "S3":
        validate_s3()

    c = Collect(save)
    if not classes:
        logger.info("Loading classes from settings")
        classes = s.classes
    if ".sql" not in classes:
        classes = list(pd.read_csv(classes, header=None).iloc[:, 0])
        logger.info(f"Found {len(classes)} classes")
    else:
        logger.info("Received a SQL for getting the classes")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(c.run(classes))
    logger.info("Successful collect command")


@cli.command(help="Collect files from CAS.")
@click.argument("id")
@click.option(
    "--settings",
    default=None,
    help="Path for settings file, if not passed default is settings.toml",
)
@click.option(
    "--secrets",
    default=None,
    help="Path for .secrets file, if not passed default is .secrets.toml",
)
@click.option(
    "--save",
    type=click.Choice(["S3", "local"], case_sensitive=False),
    default="local",
    help="Choose where to save files. If not passed default is local",
)
def cas(settings, secrets, save, id):
    load_sett_files(settings, secrets)

    validate_cas()
    cas = CAS(save=save, **s.cas)

    if save == "S3":
        validate_s3()

    if ".csv" in id:
        id = list(pd.read_csv(id).itertuples(index=False, name=None))
        logger.info(f"Got a list of {len(id)} CAS IDs")

    loop = asyncio.get_event_loop()
    loop.run_until_complete(cas.run(id))
    logger.info("Successful cas command")


@cli.command(help="Check configuration files")
@click.option(
    "--settings",
    default=None,
    help="Path for settings file, if not passed default is settings.toml",
)
@click.option(
    "--secrets",
    default=None,
    help="Path for .secrets file, if not passed default is .secrets.toml",
)
@click.option("--s3", is_flag=True)
@click.option("--db", is_flag=True)
@click.option("--cas", is_flag=True)
def check(settings, secrets, s3, db, cas):
    if not any([s3, db, cas]):
        check_all()
    else:
        check_all(s3, db, cas)


def main():
    cli()



from .db2 import DB2
from .oracle import Oracle
from .sqlserver import SQLServer

SUPPORTED_DBs = {"Oracle": Oracle, "DB2": DB2, "SQLServer": SQLServer}


def get_DB_client(DBM, *args, **kwargs):
    return SUPPORTED_DBs[DBM](*args, **kwargs)



import logging

import cx_Oracle

from .abstract import Abstract_DB
from soft_collect.config import settings as s

logger = logging.getLogger(__name__)


class Oracle(Abstract_DB):
    def select(self, sql):
        with self.set_up_connection() as conn:
            c = conn.cursor()
            yield from c.execute(sql)

    def fetch_all(self, sql):
        with self.set_up_connection() as conn:
            c = conn.cursor()
            return c.execute(sql).fetchall()

    def set_up_connection(self, _retry=2):
        logger.info(f"Creating connection to SQL Server in {s.ip}")

        try:
            connection = cx_Oracle.connect(
                s.user,
                s.password,
                f"{s.ip}/{s.base}",
                encoding="UTF-8",
                nencoding="UTF-8",
            )
        except cx_Oracle.DatabaseError as e:
            logger.error(f"Can't set up a connection with {s.ip}/{s.base}, {e}")
            if _retry:
                return self.set_up_connection(_retry - 1)
            raise e

        return connection



import logging
from abc import ABC
from abc import abstractmethod

from pandas.errors import EmptyDataError

from soft_collect.config import settings

logger = logging.getLogger(__name__)


class AbstractSaveHandler(ABC):
    def __init__(self):
        s = settings
        self.dir = f"{s.alias}/{s.grau}/{s.base}"

    @abstractmethod
    def write_file(self, dirname, filename, obj):
        pass

    @abstractmethod
    def exists(self, filepath):
        pass

    @abstractmethod
    def get_file(self, filepath):
        pass

    def save_obj(self, obj, cdclass, key, part):
        pdf_dir = f"{self.dir}/{cdclass}/"  # TODO Check if pdf
        filename = f"{key}_1/{part}.pdf"
        logger.info(f"Saving {filename}")
        self.write_file(pdf_dir, filename, obj)

    def save_meta(self, df, filename):
        meta_dir = f"{self.dir}/meta/"
        filename = f"{filename}.csv"
        obj = df.to_csv(index=False)
        logger.info(f"Saving {filename}")
        self.write_file(meta_dir, filename, obj)

    def check_obj(self, cdclass, key):
        filepath = f"{self.dir}/{cdclass}/{key}"
        logger.info(f"Checking if {filepath} exists")
        return self.exists(filepath)

    def check_meta(self, filename, return_obj=True):
        filepath = f"{self.dir}/meta/{filename}.csv"
        logger.info(f"Checking if {filepath} exists")
        res = self.exists(filepath)
        if return_obj:
            try:
                obj = self.get_file(filepath) if res else None
            except EmptyDataError:
                logger.error(f"The file {filepath} is empty")
                res = False
                obj = None
            return (res, obj)
        return res



from .local import LocalSaveHandler
from .s3 import S3SaveHandler

SUPPORTED_SAVES = {"local": LocalSaveHandler, "S3": S3SaveHandler}


def get_save_client(where, *args, **kwargs):
    return SUPPORTED_SAVES[where](*args, **kwargs)



import logging

from dynaconf import Dynaconf
from dynaconf import Validator

root = logging.getLogger()
filename = "soft-collect.log"
file_handler = logging.FileHandler(filename=filename)

formatter = logging.Formatter(
    "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
)

file_handler.setFormatter(formatter)

root.setLevel(logging.INFO)
root.addHandler(file_handler)
root.propagate = False

logger = logging.getLogger(__name__)

settings = Dynaconf(
    envvar_prefix="DYNACONF", settings_files=["settings.toml", ".secrets.toml"],
)


def validate_db():
    logger.info("Validating DB variables")
    v = [
        Validator("DBM", must_exist=True),
        Validator("IP", must_exist=True),
        Validator("BASE", must_exist=True),
    ]
    validate(v)


def validate_collect():
    logger.info("Validating Collect variables")
    v = [
        Validator("ALIAS", must_exist=True),
        Validator("GRAU", len_eq=2, must_exist=True),
        Validator("OBJS_SQL", "KEYS_SQL", must_exist=True),
    ]
    validate(v)


def validate_s3():
    logger.info("Validating S3 variables")
    v = [
        Validator("ACCESS_KEY", must_exist=True),
        Validator("SECRET_ACCESS_KEY", must_exist=True),
        Validator("BUCKET", must_exist=True),
    ]
    validate(v)


def validate_cas():
    logger.info("Validating CAS variables")
    v = [Validator("CAS", must_exist=True)]
    validate(v)


def validate(v):
    settings.validators.register(*v)
    settings.validators.validate()



import asyncio
import functools
import io
import logging
from concurrent.futures import ThreadPoolExecutor

import pandas as pd
from rich.progress import Progress

from .cas import CAS
from .config import settings
from .db import get_DB_client
from .models.objects import ObjResult
from .save import get_save_client

logger = logging.getLogger(__name__)


class Collect:
    def __init__(self, save):
        self.sh = get_save_client(save)
        self.db = get_DB_client(settings.dbm)

        if settings.get("CAS", None):
            self.cas = CAS(save=save, **settings.cas)
        else:
            self.cas = None

    async def run(self, classes):
        if isinstance(classes, str):
            classes = [row for row in self.get_rows(level="CLASSES", sql=classes)]

        loop = asyncio.get_event_loop()
        executor = ThreadPoolExecutor(200)
        with Progress() as p:
            class_task = p.add_task("[red]CLASSES", total=len(classes))
            await asyncio.gather(
                *[
                    loop.run_in_executor(
                        executor,
                        functools.partial(
                            self.parallel_classes, p, class_task, cdclass
                        ),
                    )
                    for cdclass in classes
                ]
            )

    def parallel_classes(self, progress, task, cdclass):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        executor = ThreadPoolExecutor(200)

        logger.info(f"Getting documents for class {cdclass}")
        parallel_docs = asyncio.gather(
            *[
                loop.run_in_executor(
                    executor, functools.partial(self.parallel_docs, doc, cdclass),
                )
                for doc in self.get_rows(cdclass)
            ]
        )
        loop.run_until_complete(parallel_docs)
        logger.info(f"Finished class {cdclass}")
        progress.advance(task)

    def parallel_docs(self, doc, cdclass):
        logger.info(f"Getting key {doc} in class {cdclass}")
        if doc:
            self.get_objects(doc, cdclass)
        else:
            logger.error(f"Empty key in class {cdclass}, it will be ignored")

    def get_objects(self, key, cdclass):
        if self.sh.check_obj(cdclass, key):
            logger.info(f"Object {key} already exists")
            return True
        for row in self.db.get_objects(key):
            result = ObjResult(*row)
            if result.idcas and self.cas:
                logger.info(
                    f"CAS: id={result.idcas.strip()}, {cdclass}, {result.key}, {result.part}"
                )
                # loop = asyncio.get_event_loop()
                # result.obj = loop.run_until_complete(
                #     self.cas.retrieve_obj(result.idcas)
                # )
            if result.obj:
                self.sh.save_obj(
                    io.BytesIO(result.obj).read(), cdclass, result.key, result.part
                )
            else:
                logger.error(
                    f"The key {result.key} has a empty file in part {result.part}, it will be ignored"
                )

    def get_rows(self, cdclass=None, level="DOCS", sql=None):
        if not sql:
            sql = settings.keys_sql
        if not cdclass:
            cdclass = sql
        exists, docs = self.sh.check_meta(f"{level}_LIST_{cdclass}")
        if exists:
            logger.info(f"List of rows for {cdclass} already exists")
            docs.dropna(inplace=True)
            yield from [doc[0] for _, doc in docs.iterrows()]
        else:
            rows = self.db.get_rows(sql, format={"CLASSE": cdclass})
            rows = [row[0] for row in rows]
            df = pd.DataFrame(rows).dropna()
            self.sh.save_meta(df, f"{level}_LIST_{cdclass}")
            yield from rows



import logging
from abc import ABC
from abc import abstractmethod

from soft_collect.config import settings

logger = logging.getLogger(__name__)


def get_sql(sql):
    if sql[-4:] == ".sql":
        with open(sql, encoding="utf8", errors="ignore") as f:
            return f.read()
    return sql


class Abstract_DB(ABC):
    def __init__(self):
        print("Connecting to DB...")
        self.set_up_connection()
        print("Connected to DB")

    @abstractmethod
    def set_up_connection(self, _retry=2):
        pass

    @abstractmethod
    def select(self, sql):
        pass

    @abstractmethod
    def fetch_all(self, sql):
        pass

    def get_rows(self, query, format):
        logger.info(f"Getting rows {format}")
        sql = get_sql(query).format(**format)
        return self.fetch_all(sql)

    def get_objects(self, key):
        logger.info(f"Getting objects for key {key}")
        sql = get_sql(settings.objs_sql).format(KEY=key)
        yield from self.select(sql)



import logging

import pandas as pd

from .config import settings
from .db import get_DB_client
from .save import get_save_client
from .utils import elapsed_time

logger = logging.getLogger(__name__)


class Query:
    def __init__(self, save):
        self.db = get_DB_client(settings.dbm)
        self.sh = get_save_client(save)

    def run(self, sql, filename):
        logger.info(f"Reading SQL file in {sql}")

        with open(sql, "r", encoding="utf8", errors="ignore") as f:
            query = f.read()

        logger.info("Running query")
        with elapsed_time as et:
            task = et.add_task("Running your query, elapsed time:")
            cnxn = self.db.set_up_connection()
            df_chunks = pd.read_sql(query, cnxn, coerce_float=False, chunksize=20000)
            et.stop_task(task)

        logger.info("Finished running query")

        print("Saving result in chunks of 20k rows")
        prefix = ""
        for i, df in enumerate(df_chunks):
            print(f"Saving chunk {i+1}...")
            if i > 0:
                prefix = f"{i}-"
            print(f"Saving result to {prefix+filename}.csv")
            logger.info(f"Saving result to {prefix+filename}.csv")
            self.sh.save_meta(df, prefix + filename)



from rich.console import Console
from rich.progress import Progress
from rich.table import Table


console = Console()

progress_bar = Progress(
    "[progress.description]{task.description}",
    "[yellow]Completed: {task.completed}",
    "[yellow]Elapsed Time: {task.elapsed:5.1f}s ",
    "[blue]Speed: {task.speed} it/s ",
)

elapsed_time = Progress(
    "[progress.description]{task.description}", "[yellow] {task.elapsed:5.1f}s ",
)


def print_table(df, title):
    table = Table(*[str(col) for col in df.columns], title=title)
    for _, row in df.iterrows():
        table.add_row(*[str(cell) for cell in row])
    console.print(table)



import io
import logging

import botocore
import pandas as pd
from tucuxi import Session

from .abstract import AbstractSaveHandler
from soft_collect.config import settings

logger = logging.getLogger(__name__)


class S3SaveHandler(AbstractSaveHandler):
    def __init__(self):
        super().__init__()
        logger.info(f"Connecting to AWS S3 Bucket {settings.bucket}")
        proxy = settings.get("proxy", None)
        if proxy:
            logger.info(f"Using proxy for S3 connection. Proxy: {settings.proxy}")
        sess = Session(settings.access_key, settings.secret_access_key,)
        self.s3 = sess.s3(
            settings.bucket,
            config=botocore.client.Config(proxies=proxy, max_pool_connections=1000),
        )

    def write_file(self, dirname, filename, obj):
        key = dirname + filename
        self.s3.set_object(key, obj)

    def exists(self, prefix):
        return bool(list(self.s3.list_objects(prefix)))

    def get_file(self, filepath):
        if filepath[-3:] == "csv":
            csv = io.BytesIO(self.s3.get_object(filepath))
            return pd.read_csv(csv)
        return {}



import logging

import pyodbc

from .abstract import Abstract_DB
from soft_collect.config import settings as s

logger = logging.getLogger(__name__)


def get_default_driver():
    drivers = [dri for dri in pyodbc.drivers() if "SQL Server" in dri]
    return drivers[0]


class SQLServer(Abstract_DB):
    def select(self, sql):
        with self.set_up_connection() as conn:
            c = conn.cursor()
            c.execute(sql)
            row = c.fetchone()
            while row:
                yield row
                row = c.fetchone()

    def fetch_all(self, sql):
        with self.set_up_connection() as conn:
            c = conn.cursor()
            return c.execute(sql).fetchall()

    def set_up_connection(self, _retry=2):
        logger.info(f"Creating connection to SQL Server in {s.ip}")

        if not pyodbc.drivers():
            raise Exception("No SQL Server Driver found, please install it.")
        driver = "{" + s.get("DRIVER", get_default_driver()) + "}"

        conn_str = s.get("CONN_STR", None)

        try:
            connection = pyodbc.connect(
                conn_str,
                driver=driver,
                server=s.ip,
                port=s.get("PORT", None),
                database=s.base,
                user=s.get("USER", None),
                password=s.get("PASSWORD", None),
            )
        except BaseException as e:
            logger.error(f"Can't set up a connection with {s.ip}/{s.base}, {e}")
            if _retry:
                return self.set_up_connection(_retry - 1)
            raise e

        return connection



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['soft_collect',
 'soft_collect.db',
 'soft_collect.models',
 'soft_collect.save',
 'soft_collect.templates',
 'soft_collect.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'cx_Oracle>=8.0.0,<9.0.0',
 'dynaconf>=3.0.0,<4.0.0',
 'httpx>=0.17.0,<0.18.0',
 'pandas>=1.2.1,<2.0.0',
 'pyodbc>=4.0.30,<5.0.0',
 'rich>=3.0.0,<4.0.0',
 'tucuxi>=0.1.2,<0.2.0']

entry_points = \
{'console_scripts': ['soft-collect = soft_collect.console:main']}

setup_kwargs = {
    'name': 'soft-collect',
    'version': '0.2.9',
    'description': 'Soft-Collect',
    'long_description': '\nsoft-collect\n============\n\n\n\n|PyPI|  |Black| |pre-commit|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/soft-collect.svg\n   :target: https://pypi.org/project/soft-collect/\n   :alt: PyPI\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n\n\nFeatures\n--------\n\n* TODO\n\n\nRequirements\n------------\n\n* TODO\n\n\nInstallation\n------------\n\nYou can install *soft-collect* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install soft-collect\n\n\nUsage\n-----\n\n* TODO\n\n\n\n.. _PyPI: https://pypi.org/\n.. _pip: https://pip.pypa.io/\n',
    'author': 'Luccas Quadros',
    'author_email': 'luccas.quadros@softplan.com.br',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/unj-inovacao/soft-collect',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import logging
from glob import glob
from pathlib import Path

import pandas as pd

from .abstract import AbstractSaveHandler

logger = logging.getLogger(__name__)


def make_dir(path):
    path = path.rsplit("/", 1)[0]
    Path(path).mkdir(parents=True, exist_ok=True)


class LocalSaveHandler(AbstractSaveHandler):
    def __init__(self):
        super().__init__()
        logger.info("Using local to save results")
        self.dir = "data/" + self.dir

    def write_file(self, dirname, filename, obj):
        make_dir(dirname + filename)
        mode = "w" if isinstance(obj, str) else "wb"
        with open(dirname + filename, mode) as f:
            f.write(obj)

    def exists(self, filepath):
        return len(glob(f"{filepath}*")) > 0

    def get_file(self, filepath):
        if filepath[-3:] == "csv":
            return pd.read_csv(filepath)
        return {}



import asyncio
import logging
import socket
from typing import Union

import httpx

from .save import get_save_client


logger = logging.getLogger(__name__)
# TODO Print progress and change the saving path


class CAS:
    def __init__(
        self,
        save,
        host,
        port,
        username,
        password,
        protocol="http",
        context="/spsmb/ibm/ia",
    ) -> None:
        self.host = host
        self.port = port
        self.BASE_URL = f"{protocol}://{host}:{port}{context}"
        auth = (username, password)
        headers = {
            "content-type": "application/octet-stream",
            "User-Agent": "Softplan Broker Client v1.0",
        }
        limits = httpx.Limits(max_keepalive_connections=5, max_connections=10)
        self.client = httpx.AsyncClient(
            auth=auth, headers=headers, base_url=self.BASE_URL, limits=limits
        )

        self.sh = get_save_client(save)

    def head(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(1)

        try:
            s.connect((self.host, int(self.port)))
            s.shutdown(socket.SHUT_RD)
            return True

        # Connection Timed Out
        except socket.timeout:
            print("Connection timed out!")
            return False
        except OSError as e:
            print("OS Error:", e)
            return False

    async def retrieve_obj(self, id: str) -> bytes:
        RETRIEVE_OP = "/doRetrieve"

        async with self.client as client:
            logger.info(f"Requesting for file in CAS id={id}")
            res = await client.get(f"{RETRIEVE_OP}?id={id}")

        return res.content

    async def retrive_list_objs(self, list_ids: list) -> list:
        task_list = []
        for id, classe, key, part in list_ids:
            task_list.append(self.save_obj(id, classe, key, part))

        return await asyncio.gather(*task_list)

    async def save_obj(self, id, classe, key, part):
        obj = await self.retrieve_obj(id)
        self.sh.save_obj(obj, classe, key, part)

    async def run(self, ids: Union[str, list]):
        exec_function = {str: self.save_obj, list: self.retrive_list_objs}

        await exec_function[type(ids)](ids)



# TODO Use this file to validate S3, DB and CAS Connection
from .cas import CAS
from .config import settings
from .db import get_DB_client
from .save import get_save_client


def check_s3():
    print("Checking S3 connection...")
    get_save_client("s3")


def check_db():
    print("Checking DB connection...")
    get_DB_client(settings.dbm)


def check_cas():
    print("Checking CAS connection")
    CAS(save="local", **settings.cas).head()


def check_all(s3=True, db=True, cas=True):
    if s3:
        check_s3()
    if db:
        check_db()
    if cas:
        check_cas()
